"""Code for biosynonyms."""

from .model import (
    Synonym,
    SynonymTuple,
    grounder_from_synonyms,
    group_synonyms,
    parse_synonyms,
    write_synonyms,
)
from .resources import (
    get_gilda_terms,
    get_grounder,
    get_negative_synonyms,
    get_positive_synonyms,
    load_unentities,
)

__all__ = [
    "Synonym",
    "SynonymTuple",
    "get_gilda_terms",
    "get_grounder",
    "get_negative_synonyms",
    "get_positive_synonyms",
    "grounder_from_synonyms",
    "group_synonyms",
    "load_unentities",
    "parse_synonyms",
    "write_synonyms",
]



"""Sort the synonyms file."""

from biosynonyms.model import lint_synonyms

from .resources import (
    NEGATIVES_PATH,
    POSITIVES_PATH,
    _load_unentities,
    write_unentities,
)


def _main() -> None:
    lint_synonyms(POSITIVES_PATH)
    lint_synonyms(NEGATIVES_PATH)
    write_unentities(list(_load_unentities()))


if __name__ == "__main__":
    _main()



"""Predict synonyms based on similarity.

Does the following:

- [ ] Automate acquisition of INDRA DB processed statements
- [x] Convert processed (including ungrounded statements) into triples
- [ ] Calculate graph embedding
- [ ] Calculate nearest neighbors for top K entities
      with text to all entities with grounding

Run with ``python -m biosynonyms.predict``
"""

import gzip
import itertools as itt
import json
import logging
from collections import Counter
from collections.abc import Iterable
from functools import partial
from itertools import permutations
from pathlib import Path
from typing import TYPE_CHECKING

import bioregistry
import click
import gilda
import pandas as pd
import pystow
from curies import ReferenceTuple
from indra.assemblers.indranet.assembler import NS_PRIORITY_LIST
from indra.statements import (
    Agent,
    Association,
    Complex,
    Conversion,
    Influence,
    Statement,
)
from more_click import force_option
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from biosynonyms.resources import load_unentities

if TYPE_CHECKING:
    import ensmallen

logger = logging.getLogger(__name__)

MODULE = pystow.module("indra", "db")
PAIRS_PATH = MODULE.join(name="biosynonyms_pairs.tsv")
COUNTER_PATH = MODULE.join(name="biosynonyms_counter.tsv")
COUNTER_TOP_PATH = MODULE.join(name="biosynonyms_counter_top_1000.tsv")
EMBEDDINGS_PATH = MODULE.join(name="biosynonyms_embeddings.parquet")
PLOT_PATH = MODULE.join(name="plot.png")
TEXT_PREFIX = "text"

Row = tuple[ReferenceTuple, ReferenceTuple]
Rows = list[Row]


def ensure_procesed_statements() -> Path:
    """Ensure the latest processed INDRA statements file is downloaded from S3."""
    # s3://bigmech/indra-db/dumps/principal/2023-05-05/processed_statements.tsv.gz
    bucket = "bigmech"
    key = "indra-db/dumps/principal/2023-05-05/processed_statements.tsv.gz"
    return MODULE.ensure_from_s3("principal", "2023-05-05", s3_bucket=bucket, s3_key=key)


def norm(s: str) -> str:
    """Normalize a string."""
    return s.strip().replace("\t", " ").replace("\n", " ").replace("  ", " ")


def _norm_strict(prefix: str, identifier: str) -> ReferenceTuple:
    norm_prefix, norm_identifier = bioregistry.normalize_parsed_curie(prefix, identifier)  # type:ignore [attr-defined]
    if not norm_prefix or not norm_identifier:
        raise ValueError
    return ReferenceTuple(norm_prefix, norm_identifier)


def get_agent_curie_tuple(agent: Agent, *, grounder: gilda.Grounder) -> ReferenceTuple:
    """Return a tuple of name space, id from an Agent's db_refs."""
    for prefix in NS_PRIORITY_LIST:
        if prefix in agent.db_refs:
            return _norm_strict(prefix, agent.db_refs[prefix])

    norm_agent_name = norm(agent.name)
    scored_matches = grounder.ground(norm_agent_name)
    if not scored_matches:
        return ReferenceTuple(TEXT_PREFIX, norm_agent_name)

    scored_match = scored_matches[0]
    return _norm_strict(scored_match.term.db, scored_match.term.id)


@click.command()
@click.option("--size", type=int, default=32)
@force_option  # type:ignore
def main(size: int, force: bool) -> None:
    """Generate synonym predictions."""
    if not EMBEDDINGS_PATH.is_file() or force:
        graph = get_graph(force=force)
        graph = graph.remove_disconnected_nodes()

        from embiggen.embedders.ensmallen_embedders.second_order_line import (
            SecondOrderLINEEnsmallen,
        )

        click.echo("Fitting Second Order LINE")
        embedding = SecondOrderLINEEnsmallen(embedding_size=size).fit_transform(graph)
        df: pd.DataFrame = embedding.get_all_node_embedding()[0].sort_index()
        df.index.name = "node"
        df.columns = [str(c) for c in df.columns]
        click.echo(f"Writing Parquet to {EMBEDDINGS_PATH}")
        df.to_parquet(EMBEDDINGS_PATH)
        # TODO output index of all synonyms
        # TODO calculate closest neighbors for synonyms
        #  (that aren't already in predictions)

        import matplotlib.pyplot as plt
        from embiggen import GraphVisualizer

        visualizer = GraphVisualizer(graph)
        fig, _axes = visualizer.fit_and_plot_all(embedding)
        click.echo(f"Outputting plots to {PLOT_PATH}")
        plt.savefig(PLOT_PATH, dpi=300)
        plt.close(fig)


def get_grounder() -> gilda.Grounder:
    """Get a Gilda grounder."""
    # TODO get both gilda built in plus extras from positives
    raise NotImplementedError


def get_graph(force: bool = False, *, multiprocessing: bool = False) -> "ensmallen.Graph":
    """Get a the undirected INDRA graph."""
    if not PAIRS_PATH.exists() or force:
        click.echo("loading non-entities")
        unentities = load_unentities()

        click.echo("Getting up Gilda grounder")
        grounder = get_grounder()
        click.echo("Warming up grounder")
        grounder.ground("k-ras")

        func = partial(_line_to_rows, unentities=unentities, grounder=grounder)

        click.echo("Ensuring INDRA statements from S3")
        input_path = ensure_procesed_statements()
        tqdm_kwargs = {
            "desc": "loading INDRA db",
            "unit": "statement",
            "unit_scale": True,
            "total": 65_102_088,
        }
        click.echo("Reading INDRA statements")
        with gzip.open(input_path, "rt") as file:
            click.echo(f"Opened {file.name}")

            if multiprocessing:
                groups = process_map(
                    func,
                    file,
                    **tqdm_kwargs,
                    max_workers=4,
                    chunksize=300_000,
                )
                rows: set[Row] = set(itt.chain.from_iterable(groups))
            else:
                rows = set(
                    itt.chain.from_iterable(func(line) for line in tqdm(file, **tqdm_kwargs))
                )

        sorted_rows = sorted(rows)

        click.echo("Tabulating entity counts")
        counter = Counter(_iter_names_from_rows(sorted_rows))
        counter_df = pd.DataFrame(counter.most_common(), columns=["synonym", "count"])
        counter_df.to_csv(COUNTER_PATH, sep="\t", index=False)
        counter_df.head(1000).to_csv(COUNTER_TOP_PATH, sep="\t", index=False)

        click.echo(f"Writing graph to {PAIRS_PATH}")
        # this can't be gzipped or else GRAPE doesn't work
        with PAIRS_PATH.open("w") as file:
            for source, target in tqdm(sorted_rows, desc="writing", unit_scale=True):
                print(
                    source.curie,
                    target.curie,
                    sep="\t",
                    file=file,
                )

    from ensmallen import Graph

    click.echo(f"Loading graph from {PAIRS_PATH}")
    return Graph.from_csv(
        edge_path=str(PAIRS_PATH),
        edge_list_separator="\t",
        sources_column_number=0,
        destinations_column_number=1,
        edge_list_numeric_node_ids=False,
        directed=True,
        name="INDRA Database",
        verbose=True,
    )


def _iter_names_from_rows(sorted_rows: Iterable[Row]) -> Iterable[str]:
    it = tqdm(sorted_rows, unit_scale=True, desc="counting occurrences")
    for source, target in it:
        if source.prefix == TEXT_PREFIX:
            yield source.identifier
        if target.prefix == TEXT_PREFIX:
            yield target.identifier


def _line_to_rows(line: str, unentities: set[str], grounder: gilda.Grounder) -> Rows:
    _assembled_hash, stmt_json_str = line.split("\t", 1)
    # why won't it strip the extra?!?!
    stmt_json_str = stmt_json_str.replace('""', '"').strip('"')[:-2]
    stmt = Statement._from_json(json.loads(stmt_json_str))
    return _rows_from_stmt(stmt, unentities=unentities, grounder=grounder)


def _rows_from_stmt(  # noqa:C901
    stmt: Statement,
    *,
    unentities: set[str],
    grounder: gilda.Grounder,
    complex_members: int = 3,
) -> Rows:
    not_none_agents = stmt.real_agent_list()
    if len(not_none_agents) < 2:
        # Exclude statements with less than 2 agents
        return []

    if isinstance(stmt, Influence | Association):
        # Special handling for Influences and Associations
        stmt_pol = stmt.overall_polarity()
        if stmt_pol == 1:
            sign = 0
        elif stmt_pol == -1:
            sign = 1
        else:
            sign = None
        if isinstance(stmt, Influence):
            edges = [(stmt.subj.concept, stmt.obj.concept, sign)]
        else:
            edges = [(a, b, sign) for a, b in permutations(not_none_agents, 2)]
    elif isinstance(stmt, Complex):
        # Handle complexes by creating pairs of their
        # not-none-agents.

        # Do not add complexes with more members than complex_members
        if len(not_none_agents) > complex_members:
            logger.debug(f"Skipping a complex with {len(not_none_agents)} members.")
            return []
        else:
            # add every permutation with a neutral polarity
            edges = [(a, b, None) for a, b in permutations(not_none_agents, 2)]
    elif isinstance(stmt, Conversion):
        edges = []
        if stmt.subj:
            for obj in stmt.obj_from:
                edges.append((stmt.subj, obj, 1))
            for obj in stmt.obj_to:
                edges.append((stmt.subj, obj, 0))
    elif len(not_none_agents) > 2:
        # This is for any remaining statement type that may not be
        # handled above explicitly but somehow has more than two
        # not-none-agents at this point
        return []
    else:
        edges = [(not_none_agents[0], not_none_agents[1], None)]

    def _is_unentity(r: ReferenceTuple) -> bool:
        return r.prefix == TEXT_PREFIX and r.identifier in unentities

    rows = []
    for agent_a, agent_b, _sign in edges:
        if agent_a.name == agent_b.name:
            continue
        source = get_agent_curie_tuple(agent_a, grounder=grounder)
        target = get_agent_curie_tuple(agent_b, grounder=grounder)
        if _is_unentity(source) or _is_unentity(target):
            continue

        row = (source, target)
        rows.append(row)
    return rows


# TODO open up pairs file and clean it to remove any row where
#  either the source or target are in the unentities list


if __name__ == "__main__":
    main()



"""A data model for synonyms."""

from __future__ import annotations

import csv
import datetime
import importlib.util
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, NamedTuple, TypeAlias

import requests
from curies import NamedReference, Reference
from curies import vocabulary as v
from pydantic import BaseModel, Field
from pydantic_extra_types.language_code import LanguageAlpha2
from tqdm import tqdm

if TYPE_CHECKING:
    import gilda

__all__ = [
    "Synonym",
    "SynonymTuple",
    "append_synonym",
    "group_synonyms",
    "lint_synonyms",
    "parse_synonyms",
    "write_synonyms",
]


class SynonymTuple(NamedTuple):
    """Represents rows in a spreadsheet."""

    text: str
    curie: str
    name: str
    predicate: str
    type: str | None
    provenance: str | None
    contributor: str | None
    date: str | None
    language: str | None
    comment: str | None
    source: str | None


#: The header for the spreadsheet
HEADER = list(SynonymTuple._fields)

#: A set of permissible predicates
PREDICATES = [v.has_label, *v.synonym_scopes.values()]


class Synonym(BaseModel):
    """A data model for synonyms."""

    text: str
    language: LanguageAlpha2 | None = Field(
        None,
        description="The language of the synonym. If not given, typically "
        "assumed to be american english.",
    )
    reference: NamedReference
    predicate: Reference = Field(
        default=v.has_related_synonym,
        description="The predicate that connects the term (as subject) "
        "to the textual synonym (as object)",
        examples=PREDICATES,
    )
    type: Reference | None = Field(
        default=None,
        title="Synonym type",
        description="See the OBO Metadata Ontology for valid values",
        examples=list(v.synonym_types),
    )

    provenance: list[Reference] = Field(
        default_factory=list,
        description="A list of articles (e.g., from PubMed, PMC, arXiv) where this synonym appears",
    )
    contributor: Reference | None = Field(
        None,
        description="The contributor, usually given as a reference to ORCID",
        examples=[v.charlie],
    )
    comment: str | None = Field(
        None, description="An optional comment on the synonym curation or status"
    )
    source: str | None = Field(
        None, description="The name of the resource where the synonym was curated"
    )
    date: datetime.datetime | None = Field(None, description="The date of initial curation")

    def get_all_references(self) -> set[Reference]:
        """Get all references made by this object."""
        rv: set[Reference] = {self.reference, self.predicate, *self.provenance}
        if self.type:
            rv.add(self.type)
        if self.contributor:
            rv.add(self.contributor)
        return rv

    @property
    def name(self) -> str:
        """Get the reference's name."""
        return self.reference.name

    @property
    def curie(self) -> str:
        """Get the reference's CURIE."""
        return self.reference.curie

    @property
    def date_str(self) -> str:
        """Get the date as a string."""
        if self.date is None:
            raise ValueError("date is not set")
        return self.date.strftime("%Y-%m-%d")

    @classmethod
    def from_row(
        cls, row: dict[str, Any], *, names: Mapping[Reference, str] | None = None
    ) -> Synonym:
        """Parse a dictionary representing a row in a TSV."""
        reference = Reference.from_curie(row["curie"])
        name = (names or {}).get(reference) or row.get("name") or row["text"]
        data = {
            "text": row["text"],
            "reference": NamedReference(
                prefix=reference.prefix, identifier=reference.identifier, name=name
            ),
            "predicate": (
                Reference.from_curie(predicate_curie.strip())
                if (predicate_curie := row.get("predicate"))
                else v.has_related_synonym
            ),
            "type": _safe_parse_curie(row["type"]) if "type" in row else None,
            "provenance": [
                Reference.from_curie(provenance_curie.strip())
                for provenance_curie in (row.get("provenance") or "").split(",")
                if provenance_curie.strip()
            ],
            # get("X") or None protects against empty strings
            "language": row.get("language") or None,
            "comment": row.get("comment") or None,
            "source": row.get("source") or None,
        }
        if contributor_curie := (row.get("contributor") or "").strip():
            data["contributor"] = Reference.from_curie(contributor_curie)
        if date := (row.get("date") or "").strip():
            data["date"] = datetime.datetime.strptime(date, "%Y-%m-%d")

        return cls.model_validate(data)

    def _as_row(self) -> SynonymTuple:
        """Get the synonym as a row for writing."""
        return SynonymTuple(
            text=self.text,
            curie=self.curie,
            name=self.name,
            predicate=self.predicate.curie,
            type=self.type.curie if self.type else None,
            provenance=",".join(p.curie for p in self.provenance) if self.provenance else None,
            contributor=self.contributor.curie if self.contributor is not None else None,
            date=self.date_str if self.date is not None else None,
            language=self.language or None,
            comment=self.comment or None,
            source=self.source or None,
        )

    @classmethod
    def from_gilda(cls, term: gilda.Term) -> Synonym:
        """Get this synonym as a gilda term.

        :param term: A Gilda term
        :returns: A synonym object

        .. warning::

            Gilda's data model is less complete, so resulting synonym objects
            will not have detailed curation provenance
        """
        data = {
            "text": term.text,
            # TODO standardize?
            "reference": NamedReference(prefix=term.db, identifier=term.id, name=term.entry_name),
            "source": term.source,
        }
        return cls.model_validate(data)

    def to_gilda(self, organism: str | None = None) -> gilda.Term:
        """Get this synonym as a gilda term."""
        if not self.name:
            raise ValueError("can't make a Gilda term without a label")
        return _gilda_term(
            text=self.text,
            reference=self.reference,
            name=self.name,
            status=_get_gilda_status(self),
            source=self.source,
            organism=organism,
        )


#: See https://github.com/gyorilab/gilda/blob/ea328734f26c91189438e6d3408562f990f38644/gilda/term.py#L167C1-L167C69
GildaStatus: TypeAlias = Literal["name", "synonym", "curated", "former_name"]


def _get_gilda_status(synonym: Synonym) -> GildaStatus:
    """Get the Gilda status for a synonym."""
    if synonym.predicate and synonym.predicate.pair == v.has_label.pair:
        return "name"
    if synonym.type and synonym.type.pair == v.previous_name.pair:
        return "former_name"
    return "synonym"


def _gilda_term(
    *,
    text: str,
    reference: Reference,
    name: str | None = None,
    status: GildaStatus,
    source: str | None,
    organism: str | None = None,
) -> gilda.Term:
    import gilda
    from gilda.process import normalize

    return gilda.Term(
        normalize(text),
        text=text,
        db=reference.prefix,
        id=reference.identifier,
        entry_name=name or text,
        status=status,
        source=source,
        organism=organism,
    )


def _safe_parse_curie(x) -> Reference | None:  # type:ignore
    if not isinstance(x, str) or not x.strip():
        return None
    return Reference.from_curie(x.strip())


def write_synonyms(path: str | Path, synonyms: Iterable[Synonym]) -> None:
    """Write synonyms to a path."""
    path = Path(path).expanduser().resolve()
    rows = (synonym._as_row() for synonym in synonyms)
    if importlib.util.find_spec("pandas"):
        _write_pandas(path, rows)
    else:
        _write_builtin(path, rows)


def _write_builtin(path: Path, rows: Iterable[SynonymTuple]) -> None:
    with path.open("w") as file:
        writer = csv.writer(file, delimiter="\t")
        writer.writerow(HEADER)
        writer.writerows(rows)


def _write_pandas(path: Path, rows: Iterable[SynonymTuple]) -> None:
    import pandas as pd

    df = pd.DataFrame(rows, columns=HEADER)
    remove = [col for col in HEADER if df[col].isna().all()]
    for col in remove:
        del df[col]

    df.to_csv(path, index=False, sep="\t")


def append_synonym(path: str | Path, synonym: Synonym) -> None:
    """Append a synonym to an existing file."""
    raise NotImplementedError


def parse_synonyms(
    path: str | Path,
    *,
    delimiter: str | None = None,
    names: Mapping[Reference, str] | None = None,
) -> list[Synonym]:
    """Load synonyms from a file.

    :param path: A local file path or URL for a biosynonyms-flavored CSV/TSV file
    :param delimiter: The delimiter for the CSV/TSV file. Defaults to tab
    :param names: A pre-parsed dictionary from references
        (i.e., prefix-luid pairs) to default labels
    :returns: A list of synonym objects parsed from the table
    """
    if isinstance(path, str) and any(path.startswith(schema) for schema in ("https://", "http://")):
        res = requests.get(path, timeout=15)
        res.raise_for_status()
        return _from_lines(res.iter_lines(decode_unicode=True), delimiter=delimiter, names=names)

    path = Path(path).resolve()

    if path.suffix == ".numbers":
        return _parse_numbers(path, names=names)

    with path.open() as file:
        return _from_lines(file, delimiter=delimiter, names=names)


def _parse_numbers(
    path: str | Path,
    *,
    names: Mapping[Reference, str] | None = None,
) -> list[Synonym]:
    # code example from https://pypi.org/project/numbers-parser
    import numbers_parser

    doc = numbers_parser.Document(path)
    sheets = doc.sheets
    tables = sheets[0].tables
    header, *rows = tables[0].rows(values_only=True)
    return _from_dicts((dict(zip(header, row, strict=False)) for row in rows), names=names)


def _from_lines(
    lines: Iterable[str],
    *,
    delimiter: str | None = None,
    names: Mapping[Reference, str] | None = None,
) -> list[Synonym]:
    return _from_dicts(csv.DictReader(lines, delimiter=delimiter or "\t"), names=names)


def _from_dicts(
    dicts: Iterable[dict[str, Any]],
    *,
    names: Mapping[Reference, str] | None = None,
) -> list[Synonym]:
    rv = []
    for i, record in enumerate(dicts, start=2):
        record = {k: v for k, v in record.items() if k and v and k.strip() and v.strip()}
        if record:
            try:
                synonym = Synonym.from_row(record, names=names)
            except ValueError as e:
                raise ValueError(f"failed on row {i}: {record}") from e
            rv.append(synonym)
    return rv


def group_synonyms(synonyms: Iterable[Synonym]) -> dict[Reference, list[Synonym]]:
    """Aggregate synonyms by reference."""
    dd: defaultdict[Reference, list[Synonym]] = defaultdict(list)
    for synonym in tqdm(synonyms, unit="synonym", unit_scale=True, leave=False):
        dd[synonym.reference].append(synonym)
    return dict(dd)


def grounder_from_synonyms(synonyms: Iterable[Synonym]) -> gilda.Grounder:
    """Get a Gilda grounder from synonyms."""
    import gilda

    rv = gilda.Grounder([synonym.to_gilda() for synonym in synonyms])
    return rv


def lint_synonyms(path: Path) -> None:
    """Lint a synonyms file."""
    with path.open() as file:
        header, *rows = (line.strip().split("\t") for line in file)
    rows = sorted(rows, key=_sort_key)
    with path.open("w") as file:
        print(*header, sep="\t", file=file)
        for row in rows:
            print(*row, sep="\t", file=file)


def _sort_key(row: Sequence[str]) -> tuple[str, str, str, str]:
    """Return a key for sorting a row."""
    return row[0].casefold(), row[0], row[1].casefold(), row[1]



"""Generate OWL from the positive synonyms."""

# TODO re-implement this using pyobo.

from __future__ import annotations

import gzip
from collections import ChainMap
from pathlib import Path
from textwrap import dedent
from typing import Annotated, Any, TextIO

import bioregistry
from curies import Reference
from typing_extensions import Doc

from biosynonyms.model import Synonym, group_synonyms
from biosynonyms.resources import get_positive_synonyms

HERE = Path(__file__).parent.resolve()
EXPORT = HERE.parent.parent.joinpath("exports")
EXPORT.mkdir(exist_ok=True)

TTL_PATH = EXPORT.joinpath("biosynonyms.ttl")
URI = "https://w3id.org/biopragmatics/resources/biosynonyms.ttl"

METADATA = f"""\
<{URI}> a owl:Ontology ;
    dcterms:title "Biosynonyms in OWL" ;
    dcterms:description "An ontology representation of community curated synonyms in Biosynonyms" ;
    dcterms:license <https://creativecommons.org/publicdomain/zero/1.0/> ;
    rdfs:comment "Built by https://github.com/biopragmatics/biosynonyms"^^xsd:string .
"""

PREAMBLE = """\
rdfs:label   a owl:AnnotationProperty; rdfs:label "label"^^xsd:string .
rdfs:seeAlso a owl:AnnotationProperty; rdfs:label "see also"^^xsd:string .
rdfs:comment a owl:AnnotationProperty; rdfs:label "comment"^^xsd:string .

oboInOwl:hasSynonym a owl:AnnotationProperty;
    rdfs:label "has synonym"^^xsd:string .

oboInOwl:hasExactSynonym a owl:AnnotationProperty;
    rdfs:label "has exact synonym"^^xsd:string .

oboInOwl:hasNarrowSynonym a owl:AnnotationProperty;
    rdfs:label "has narrow synonym"^^xsd:string .

oboInOwl:hasBroadSynonym a owl:AnnotationProperty;
    rdfs:label "has broad synonym"^^xsd:string .

oboInOwl:hasRelatedSynonym a owl:AnnotationProperty;
    rdfs:label "has related synonym"^^xsd:string .

oboInOwl:hasSynonymType a owl:AnnotationProperty;
    rdfs:label "has synonym type"^^xsd:string .

oboInOwl:hasDbXref a owl:AnnotationProperty;
    rdfs:label "has database cross-reference"^^xsd:string .

skos:exactMatch a owl:AnnotationProperty; rdfs:label "exact match"^^xsd:string .

dcterms:contributor a owl:AnnotationProperty; rdfs:label "contributor"^^xsd:string .
dcterms:source      a owl:AnnotationProperty; rdfs:label "source"^^xsd:string .
dcterms:license     a owl:AnnotationProperty; rdfs:label "license"^^xsd:string .
dcterms:description a owl:AnnotationProperty; rdfs:label "description"^^xsd:string .

BFO:0000051 a owl:ObjectProperty; rdfs:label "has part"^^xsd:string .

NCBITaxon:9606 a owl:Class ;
    rdfs:label "Homo sapiens" .

orcid:0000-0003-4423-4370 a NCBITaxon:9606 ;
    rdfs:label "Charles Tapley Hoyt"@en .

orcid:0000-0001-9439-5346 a NCBITaxon:9606 ;
    rdfs:label "Benjamin M. Gyori"@en .

# See new OMO synonyms at
# https://github.com/information-artifact-ontology/ontology-metadata/blob/master/src/templates/annotation_properties.tsv

OMO:0003000 a owl:AnnotationProperty;
    rdfs:label "abbreviation"^^xsd:string .

OMO:0003001 a owl:AnnotationProperty;
    rdfs:label "ambiguous synonym"^^xsd:string .

OMO:0003002 a owl:AnnotationProperty;
    rdfs:label "dubious synonym"^^xsd:string .

OMO:0003003 a owl:AnnotationProperty;
    rdfs:label "layperson synonym"^^xsd:string .

OMO:0003004 a owl:AnnotationProperty;
    rdfs:label "plural form"^^xsd:string .

OMO:0003005 a owl:AnnotationProperty;
    rdfs:label "UK spelling"^^xsd:string .

OMO:0003006 a owl:AnnotationProperty;
    rdfs:label "misspelling"^^xsd:string .

OMO:0003007 a owl:AnnotationProperty;
    rdfs:label "misnomer"^^xsd:string .

OMO:0003008 a owl:AnnotationProperty;
    rdfs:label "previous name"^^xsd:string .

OMO:0003009 a owl:AnnotationProperty;
    rdfs:label "legal name"^^xsd:string .

OMO:0003010 a owl:AnnotationProperty;
    rdfs:label "International Nonproprietary Name"^^xsd:string .

OMO:0003011 a owl:AnnotationProperty;
    rdfs:label "latin term"^^xsd:string .

OMO:0003012 a owl:AnnotationProperty;
    rdfs:label "acronym"^^xsd:string .
"""


def _text_for_turtle(synonym: Synonym) -> str:
    """Get the text ready for an object slot in Turtle, with optional language tag."""
    tt = f'"{_clean_str(synonym.text)}"'
    if synonym.language:
        tt += f"@{synonym.language}"
    return tt


def write_owl_rdf(**kwargs: Any) -> None:
    """Write OWL RDF in a Turtle file."""
    with open(TTL_PATH, "w") as file:
        _write_owl_rdf(get_positive_synonyms(), file, metadata=METADATA, **kwargs)


def write_owl_rdf_gz(**kwargs: Any) -> None:
    """Write OWL RDF in a gzipped Turtle file."""
    with gzip.open(TTL_PATH.with_suffix(".gz"), "wt") as file:
        _write_owl_rdf(get_positive_synonyms(), file, metadata=METADATA, **kwargs)


DEFAULT_PREFIXES: dict[str, str] = {
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "dcterms": "http://purl.org/dc/terms/",
    "owl": "http://www.w3.org/2002/07/owl#",
    "oboInOwl": "http://www.geneontology.org/formats/oboInOwl#",
    "skos": "http://www.w3.org/2004/02/skos/core#",
    "orcid": "https://orcid.org/",
    "OMO": "http://purl.obolibrary.org/obo/OMO_",
    "NCBITaxon": "http://purl.obolibrary.org/obo/NCBITaxon_",
    "BFO": "http://purl.obolibrary.org/obo/BFO_",
}


def _get_prefixes(dd: dict[Reference, list[Synonym]]) -> set[str]:
    return {
        reference.prefix
        for synonyms in dd.values()
        for synonym in synonyms
        for reference in synonym.get_all_references()
    }


def write_prefix_map(
    prefixes: set[str], file: TextIO, *, prefix_map: dict[str, str] | None = None
) -> None:
    """Write the prefix map to the top of a turtle file."""
    for prefix, uri_prefix in iter_prefix_map(prefixes, prefix_map=prefix_map):
        file.write(f"@prefix {prefix}: <{uri_prefix}> .\n")


def iter_prefix_map(
    prefixes: set[str],
    *,
    prefix_map: dict[str, str] | None = None,
) -> list[tuple[str, str]]:
    """Generate a prefix map."""
    looked_up_prefix_map: dict[str, str] = {}
    for prefix in prefixes:
        if prefix_map and prefix in prefix_map:
            pass  # given explicitly, no need to look up in bioregistry
        elif prefix not in looked_up_prefix_map:
            resource = bioregistry.get_resource(prefix)  # type:ignore [attr-defined]
            if resource is None:
                raise ValueError(f"Prefix {prefix} is not registered in the Bioregistry")
            uri_prefix = resource.rdf_uri_format or resource.get_uri_prefix()
            if uri_prefix is None:
                raise ValueError(
                    f"Prefix has no URI expansion in Bioregistry: "
                    f"{prefix} ({bioregistry.get_name(prefix)})"  # type:ignore [attr-defined]
                )
            looked_up_prefix_map[prefix] = uri_prefix

    chained_prefix_map = ChainMap(DEFAULT_PREFIXES, looked_up_prefix_map, prefix_map or {})
    return sorted(chained_prefix_map.items(), key=lambda i: i[0].casefold())


def get_axiom_str(reference: Reference, synonym: Synonym) -> str | None:
    """Get the axiom string for a synonym."""
    axiom_parts = []
    if synonym.contributor:
        axiom_parts.append(f"dcterms:contributor {synonym.contributor.curie}")
    if synonym.date:
        axiom_parts.append(f'dcterms:date "{synonym.date_str}"^^xsd:date')
    if synonym.source:
        axiom_parts.append(f'dcterms:source "{_clean_str(synonym.source)}"')
    if synonym.type:
        axiom_parts.append(f"oboInOwl:hasSynonymType {synonym.type.curie}")
    for rr in synonym.provenance:
        axiom_parts.append(f"oboInOwl:hasDbXref {rr.curie}")
    if synonym.comment:
        axiom_parts.append(f'rdfs:comment "{_clean_str(synonym.comment)}"')

    if not axiom_parts:
        # if there's no additional context to add, then we don't need to make an axiom
        return None

    axiom_parts_str = " ;\n".join(f"    {ax}" for ax in axiom_parts) + " ."
    axiom = f"""\
[
    a owl:Axiom ;
    owl:annotatedSource {reference.curie} ;
    owl:annotatedProperty {synonym.predicate.curie} ;
    owl:annotatedTarget {_text_for_turtle(synonym)} ;
{axiom_parts_str}
] .
"""
    return axiom


def _write_owl_rdf(  # noqa:C901
    synonyms: list[Synonym],
    file: TextIO,
    *,
    prefix_definitions: Annotated[
        bool, Doc("Should the @prefix definitions be added at the top?")
    ] = True,
    class_definitions: Annotated[bool, Doc("Should the `a owl:Class` and label be added?")] = True,
    metadata: str | None = None,
    prefix_map: dict[str, str] | None = None,
) -> None:
    dd = group_synonyms(synonyms)

    if prefix_definitions:
        write_prefix_map(_get_prefixes(dd), file=file, prefix_map=prefix_map)

    if metadata:
        file.write(f"\n{metadata}\n")

    file.write(f"\n{PREAMBLE}\n")

    for reference, synonyms in dd.items():
        mains: list[str] = []
        axiom_strs: list[str] = []
        for synonym in synonyms:
            mains.append(f"{synonym.predicate.curie} {_text_for_turtle(synonym)}")
            if axiom_str := get_axiom_str(reference, synonym):
                axiom_strs.append(axiom_str)

        if class_definitions:
            file.write(f"\n{reference.curie} a owl:Class ;\n")
            try:
                name = next(synonym.name for synonym in synonyms if synonym.name)
            except StopIteration:
                pass  # could not extract a name, no worries!
            else:
                mains.append(f'rdfs:label "{_clean_str(name)}"')
        else:
            file.write(f"\n{reference.curie} ")

        file.write(" ;\n".join(f"    {m}" for m in mains) + " .\n")
        if axiom_strs:
            file.write("\n")
        for axiom_str in axiom_strs:
            file.write(dedent(axiom_str))


def _clean_str(s: str) -> str:
    return s


if __name__ == "__main__":
    write_owl_rdf()



"""Tests for data integrity of biosynoynms database."""

import tempfile
import unittest
from collections import Counter
from pathlib import Path

import bioregistry
from curies import NamedReference, Reference, ReferenceTuple
from curies import vocabulary as v

import biosynonyms
from biosynonyms.model import Synonym, _sort_key
from biosynonyms.resources import (
    NEGATIVES_PATH,
    POSITIVES_PATH,
    SYNONYM_PREDICATE_CURIES,
    UNENTITIES_PATH,
    _unentities_key,
)


class TestIntegrity(unittest.TestCase):
    """Test case for data integrity tests."""

    def assert_curie(self, curie: str):
        """Assert a CURIE is standardized against the Bioregistry.

        :param curie: A compact uniform resource identifier
            of the form ``<prefix>:<identifier>``.
        """
        prefix, identifier = ReferenceTuple.from_curie(curie)

        norm_prefix = bioregistry.normalize_prefix(prefix)
        self.assertIsNotNone(norm_prefix)
        self.assertEqual(norm_prefix, prefix)

        pattern = bioregistry.get_pattern(prefix)
        if pattern:
            self.assertRegex(identifier, pattern)

    def test_positives(self):
        """Test each row in the positives file."""
        with POSITIVES_PATH.open() as file:
            _, *rows = (tuple(line.strip().split("\t")) for line in file)

        for row_index, row in enumerate(rows, start=1):
            with self.subTest(row=row_index):
                self.assertEqual(11, len(row))
                (
                    text,
                    curie,
                    _name,
                    predicate,
                    synonym_type,
                    references,
                    contributor_curie,
                    date,
                    _lang,
                    _comment,
                    _src,
                ) = row
                self.assertLess(1, len(text), msg="can not have 1 letter synonyms")
                self.assert_curie(curie)
                self.assertIn(predicate, SYNONYM_PREDICATE_CURIES)
                if synonym_type:
                    self.assertTrue(synonym_type.startswith("OMO:"))
                for reference in references.split(",") if references else []:
                    reference = reference.strip()
                    self.assert_curie(reference)
                self.assert_curie(contributor_curie)
                if date:
                    self.assertRegex(date, "\\d{4}-\\d{2}-\\d{2}")

        # test sorted
        self.assertEqual(sorted(rows, key=_sort_key), rows, msg="synonyms are not properly sorted")

        # test no duplicates
        c = Counter(row[:2] for row in rows)
        duplicated = {key: count for key, count in c.items() if count > 1}
        self.assertEqual(0, len(duplicated), msg=f"duplicated entries: {duplicated}")

    def test_negatives(self):
        """Test each row in the negatives file."""
        with NEGATIVES_PATH.open() as file:
            _, *rows = (tuple(line.strip().split("\t")) for line in file)

        for row_index, row in enumerate(rows, start=1):
            with self.subTest(row=row_index):
                self.assertEqual(5, len(row))
                text, curie, _name, references, contributor_curie = row
                self.assertLess(1, len(text), msg="can not have 1 letter synonyms")
                self.assert_curie(curie)
                for reference in references.split(","):
                    reference = reference.strip()
                    self.assert_curie(reference)
                self.assert_curie(contributor_curie)

        # test sorted
        self.assertEqual(
            sorted(rows, key=_sort_key),
            rows,
            msg="negative synonyms are not properly sorted",
        )

        # test no duplicates
        c = Counter(row[:2] for row in rows)
        duplicated = {key: count for key, count in c.items() if count > 1}
        self.assertEqual(0, len(duplicated), msg=f"duplicated entries: {duplicated}")

    def test_non_entities(self):
        """Test each row of the non-entities file."""
        with UNENTITIES_PATH.open() as file:
            _header, *rows = (line.strip().split("\t") for line in file)
        self.assertEqual(sorted(rows, key=_unentities_key), rows)
        for line_number, line in enumerate(rows, start=1):
            with self.subTest(line_number=line_number):
                self.assertEqual(2, len(line))
                text, orcid = line
                self.assertEqual(text.strip(), text)
                self.assertTrue(bioregistry.is_valid_identifier("orcid", orcid))

    def test_gilda(self):
        """Test getting tilda terms."""
        grounder = biosynonyms.get_grounder()
        scored_matches = grounder.ground("YAL021C")
        self.assertEqual(1, len(scored_matches))
        self.assertEqual("sgd", scored_matches[0].term.db)
        self.assertEqual("S000000019", scored_matches[0].term.id)

        reference = NamedReference.from_curie("test:1", "test")
        label = Reference.from_curie("rdfs:label")
        synonym_1 = Synonym(
            text="tests", predicate=v.has_exact_synonym, type=v.plural_form, reference=reference
        )
        gilda_term_1 = synonym_1.to_gilda()
        self.assertEqual("synonym", gilda_term_1.status)

        synonym_2 = Synonym(text="test", predicate=label, reference=reference)
        gilda_term_2 = synonym_2.to_gilda()
        self.assertEqual("name", gilda_term_2.status)

        synonym_3 = Synonym(
            text="old test",
            predicate=v.has_exact_synonym,
            reference=reference,
            type=v.previous_name,
        )
        gilda_term_3 = synonym_3.to_gilda()
        self.assertEqual("former_name", gilda_term_3.status)

    def test_model(self):
        """Test loading the data model."""
        biosynonyms.get_positive_synonyms()
        biosynonyms.get_negative_synonyms()

    def test_io_roundtrip(self) -> None:
        """Test IO roundtrip."""
        synonyms = biosynonyms.get_positive_synonyms()[:3]  # sample just a few

        with tempfile.TemporaryDirectory() as d:
            path = Path(d).joinpath("test.tsv")
            biosynonyms.write_synonyms(path, synonyms)
            reloaded_synonyms = biosynonyms.parse_synonyms(path)

        self.assertEqual(synonyms, reloaded_synonyms)



"""Resources for Biosynonyms."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, cast

from biosynonyms.model import PREDICATES, Synonym, grounder_from_synonyms, parse_synonyms

if TYPE_CHECKING:
    import gilda

__all__ = [
    "get_gilda_terms",
    "get_grounder",
    "get_negative_synonyms",
    "get_positive_synonyms",
    "load_unentities",
    "write_unentities",
]

HERE = Path(__file__).parent.resolve()
POSITIVES_PATH = HERE.joinpath("positives.tsv")
NEGATIVES_PATH = HERE.joinpath("negatives.tsv")
UNENTITIES_PATH = HERE.joinpath("unentities.tsv")

SYNONYM_PREDICATE_CURIES: set[str] = {p.curie for p in PREDICATES}


def load_unentities() -> set[str]:
    """Load all strings that are known not to be named entities."""
    return {line[0] for line in _load_unentities()}


def _load_unentities() -> Iterable[tuple[str, str]]:
    with UNENTITIES_PATH.open() as file:
        next(file)  # throw away header
        for line in file:
            yield cast(tuple[str, str], line.strip().split("\t"))


def _unentities_key(row: Sequence[str]) -> str:
    return row[0].casefold()


def write_unentities(rows: Iterable[tuple[str, str]]) -> None:
    """Write all strings that are known not to be named entities."""
    with UNENTITIES_PATH.open("w") as file:
        print("text", "curator_orcid", sep="\t", file=file)
        for row in sorted(rows, key=_unentities_key):
            print(*row, sep="\t", file=file)


def get_positive_synonyms() -> list[Synonym]:
    """Get positive synonyms curated in Biosynonyms."""
    return parse_synonyms(POSITIVES_PATH)


def get_negative_synonyms() -> list[Synonym]:
    """Get negative synonyms curated in Biosynonyms."""
    return parse_synonyms(NEGATIVES_PATH)


def get_gilda_terms() -> list[gilda.Term]:
    """Get Gilda terms for all positive synonyms."""
    return [synonym.to_gilda() for synonym in get_positive_synonyms()]


def get_grounder() -> gilda.Groudner:
    """Get a grounder from all positive synonyms."""
    return grounder_from_synonyms(get_positive_synonyms())
